from aitemplates.oai.utils.wrappers import retry_openai_api, metered
import logging
import os
from dotenv import load_dotenv

from typing import Any, Optional, List, Union

import aiolimiter
import openai
import openai.error
from aiohttp import ClientSession
from tqdm.asyncio import tqdm_asyncio

from aitemplates.oai.types.base import MessageDict, ChatMessages
from aitemplates.oai.ApiManager import ApiManager

load_dotenv()

OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")

if OPENAI_API_KEY is None:
    raise Exception("API key not found in environment variables")

openai.api_key = OPENAI_API_KEY


@retry_openai_api()
async def _throttled_acreate_chat_completion(
    messages: list[MessageDict],
    limiter: aiolimiter.AsyncLimiter,
    temperature: Optional[float] = 0,
    model: Optional[str] = "gpt-3.5-turbo",
    max_tokens: Optional[int] = None,
    top_p: Optional[float] = 1,
    n: Optional[int] = 1,
    stop: Optional[str] = None,
    presence_penalty: Optional[float] = 0,
    frequency_penalty: Optional[float] = 0,
) -> Any:
    """Create a throttled chat completion using the OpenAI API.

    This function ensures the number of requests made to the OpenAI API does not exceed a specified limit.

    Args:
        messages (list[MessageDict]): The messages to send to the chat completion.
        limiter (aiolimiter.AsyncLimiter): The limiter that manages throttling of requests.
        model (str, optional): The model to use. Defaults to "gpt-3.5-turbo".
        temperature (float, optional): The temperature to use. Defaults to 0.
        max_tokens (int, optional): The maximum tokens to use. Defaults to None.
        top_p (float, optional): The nucleus sampling probability. Defaults to 1.
        n (int, optional): The number of messages to generate. Defaults to 1.
        stop (str, optional): The sequence at which the generation will stop. Defaults to None.
        presence_penalty (float, optional): The presence penalty to use. Defaults to 0.
        frequency_penalty (float, optional): The frequency penalty to use. Defaults to 0.

    Returns:
        Any: The response from the chat completion.
    """
    async with limiter:
        return await openai.ChatCompletion.acreate(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            n=n,
            stop=stop,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
        )


async def async_create_chat_completion(
    parallel_calls: ChatMessages,
    model: str = "gpt-3.5-turbo",
    temperature: float = 0,
    max_tokens: Union[int, None] = None,
    print_every: int = False,
    keep_order: bool = False,
    requests_per_minute: int = 300,
    response_list: Optional[List[str]] = None,
    top_p: Optional[float] = 1,
    n: Optional[int] = 1,
    stop: Optional[str] = None,
    presence_penalty: Optional[float] = 0,
    frequency_penalty: Optional[float] = 0,
) -> list[str]:
    """Generate from OpenAI Chat Completion API asynchronously.

    Args:
        parallel_calls (ChatMessages): List of full prompts to generate from.
        model (str, optional): Model configuration. Defaults to "gpt-3.5-turbo".
        temperature (float, optional): Temperature to use. Defaults to 0.
        max_tokens (Union[int, None], optional): Maximum number of tokens to generate. Defaults to None.
        print_every (int, optional): Print the response after every this many calls. It does nothing if keep_order is True. Defaults to False.
        keep_order (bool, optional): If True, keeps the order of responses same as the input prompts. Defaults to False.
        requests_per_minute (int, optional): Number of requests per minute to allow. Defaults to 300.
        response_list (List[str], optional): If provided, responses are added to this list.
        top_p (float, optional): The nucleus sampling probability. Defaults to 1.
        n (int, optional): The number of messages to generate. Defaults to 1.
        stop (str, optional): The sequence at which the generation will stop. Defaults to None.
        presence_penalty (float, optional): The presence penalty to use. Defaults to 0.
        frequency_penalty (float, optional): The frequency penalty to use. Defaults to 0.

    Returns:
        list[str]: List of generated responses.
    """
    if keep_order and print_every:
        logging.warning("print_every do nothing since keep_order is True")
    api_manager = ApiManager()
    openai.aiosession.set(ClientSession())
    limiter = aiolimiter.AsyncLimiter(requests_per_minute)
    async_responses = [
        _throttled_acreate_chat_completion(
            model=model,
            messages=chat_sequence.raw(),
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            n=n,
            stop=stop,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            limiter=limiter,
        )
        for chat_sequence in parallel_calls.chat_sequences
    ]

    responses = []
    prompt_tokens = 0
    completion_tokens = 0

    if keep_order:
        # Keep the order of responses same as input prompts
        responses = await tqdm_asyncio.gather(*async_responses)

        # Update the cost associated with this response in the API manager
        for response in responses:
            responses.append(response.choices[0].message.content)
            prompt_tokens += response.usage.prompt_tokens
            completion_tokens += response.usage.completion_tokens

        api_manager.update_cost(
            prompt_tokens,
            completion_tokens,
            responses[0].model,
        )
    else:
        # Without order
        for future in tqdm_asyncio.as_completed(async_responses):
            response = await future

            # Update the cost associated with this response in the API manager
            api_manager.update_cost(
                response.usage.prompt_tokens,
                response.usage.completion_tokens,
                response.model,
            )

            response_text = response.choices[0].message.content
            responses.append(response_text)
            if print_every:
                logging.info(response_text + "\n")

            # Check if a list was passed and add the response_text to it
            if response_list is not None:
                response_list.append(response_text)

    # Close the session
    await openai.aiosession.get().close()  # type: ignore
    return responses



from typing import Union
from aitemplates.oai.types.base import ChatModelInfo, EmbeddingModelInfo

OPEN_AI_CHAT_MODELS = {
    info.name: info
    for info in [
        ChatModelInfo(
            name="gpt-3.5-turbo",
            prompt_token_cost=0.002,
            completion_token_cost=0.002,
            max_tokens=4096,
        ),
        ChatModelInfo(
            name="gpt-3.5-turbo-0301",
            prompt_token_cost=0.002,
            completion_token_cost=0.002,
            max_tokens=4096,
        ),
        ChatModelInfo(
            name="gpt-4",
            prompt_token_cost=0.03,
            completion_token_cost=0.06,
            max_tokens=8192,
        ),
        ChatModelInfo(
            name="gpt-4-0314",
            prompt_token_cost=0.03,
            completion_token_cost=0.06,
            max_tokens=8192,
        ),
        ChatModelInfo(
            name="gpt-4-32k",
            prompt_token_cost=0.06,
            completion_token_cost=0.12,
            max_tokens=32768,
        ),
        ChatModelInfo(
            name="gpt-4-32k-0314",
            prompt_token_cost=0.06,
            completion_token_cost=0.12,
            max_tokens=32768,
        ),
    ]
}

OPEN_AI_EMBEDDING_MODELS = {
    info.name: info
    for info in [
        EmbeddingModelInfo(
            name="text-embedding-ada-002",
            prompt_token_cost=0.0004,
            completion_token_cost=0.0,
            max_tokens=8191,
            embedding_dimensions=1536,
        ),
    ]
}

OPEN_AI_MODELS: dict[str, Union[ChatModelInfo, EmbeddingModelInfo]] = {
    **OPEN_AI_CHAT_MODELS,
    **OPEN_AI_EMBEDDING_MODELS,
}



from typing import Optional, Dict
import os
from dotenv import load_dotenv

import openai

from aitemplates.oai.utils.wrappers import retry_openai_api, metered
from aitemplates.oai.types.base import ChatSequence

load_dotenv()

OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")

if OPENAI_API_KEY is None:
    raise Exception("API key not found in environment variables")

openai.api_key = OPENAI_API_KEY


@metered
@retry_openai_api()
def create_chat_completion(
    messages: ChatSequence,
    model: str = "gpt-3.5-turbo",
    temperature: Optional[float] = 0,
    max_tokens: Optional[int] = None,
    top_p: Optional[float] = 1,
    n: Optional[int] = 1,
    stop: Optional[str] = None,
    presence_penalty: Optional[float] = 0,
    frequency_penalty: Optional[float] = 0,
) -> str:
    """Create a chat completion using the OpenAI API

    Args:
        messages (list[MessageDict]): The messages to send to the chat completion.
        model (str, optional): The model to use. Defaults to "gpt-3.5-turbo".
        temperature (float, optional): The temperature to use. Defaults to 0.
        max_tokens (int, optional): The maximum tokens to use. Defaults to None.
        top_p (float, optional): The nucleus sampling probability. Defaults to 1.
        n (int, optional): The number of messages to generate. Defaults to 1.
        stop (str, optional): The sequence at which the generation will stop. Defaults to None.
        presence_penalty (float, optional): The presence penalty to use. Defaults to 0.
        frequency_penalty (float, optional): The frequency penalty to use. Defaults to 0.

    Returns:
        Any: The response from the chat completion.
    """
    response = openai.ChatCompletion.create(
        model=model,
        messages=messages.raw(),
        temperature=temperature,
        max_tokens=max_tokens,
        top_p=top_p,
        n=n,
        stop=stop,
        presence_penalty=presence_penalty,
        frequency_penalty=frequency_penalty,
    )

    if n > 1:
        return response.choices
    return response.choices[0].message["content"]



import argparse
import nbformat as nbf
from typing import Optional


def create_notebook(
    filename: str, db: Optional[bool] = False, asnc: Optional[bool] = False
):
    """Creates a notebook using nbformat and populate the first 3 cells

    Args:
        filename (str): The name of the notebook
        db (bool, optional): Whether to use a vector database. Defaults to False.
        asnc (bool, optional): Whether to use async chat completion. Defaults to False.

    Returns:
        None
    """

    nb = nbf.v4.new_notebook()

    db1_content = """
from dotenv import load_dotenv
import os
import chromadb
from chromadb.utils import embedding_functions
chroma_client = chromadb.Client()

load_dotenv()

OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")

if OPENAI_API_KEY is None:
    raise Exception("API key not found in environment variables")

embedder = embedding_functions.OpenAIEmbeddingFunction(
    api_key=OPENAI_API_KEY,
    model_name="text-embedding-ada-002",
)

collection = chroma_client.create_collection(name="", embedding_function=embedder)"""

    # Create first cell
    cell1_content = f"""\
from aitemplates.oai.responses. {"async_chat_response import async_create_chat_completion" if asnc else "chat_response import create_chat_completion"}
from aitemplates.oai.types.base import ChatSequence, Message{", ChatMessages" if asnc else ""}
{db1_content if db else ""}
    """
    cell1 = nbf.v4.new_code_cell(cell1_content)

    # Create second cell
    cell2_content = f"""system_prompt=""
system_prompt_msg = Message("system", system_prompt)

description = ""
description_msg = Message("system", description)

user_query=""
user_query_msg = Message("user", user_query)


chat_sequence = ChatSequence([system_prompt_msg, description_msg, user_query_msg])
{"chat_sequence2 = ChatSequence([description_msg, user_query_msg])" if asnc else ""}"""

    cell2 = nbf.v4.new_code_cell(cell2_content)

    db3_content = f"""
collection.add(
    documents={"async_response1" if asnc else "[response1]"},
    metadatas=[{{"": ""}}],
    ids=[""]
)"""

    cell3_content = f"""\
response_list = []
{"async_response1 = await async_create_chat_completion(ChatMessages([chat_sequence, chat_sequence2]), print_every=True, response_list=response_list)" if asnc else "response1 = create_chat_completion(chat_sequence)"}
{db3_content if db else ""}
{"async_response1" if asnc else "response1"}"""

    cell3 = nbf.v4.new_code_cell(cell3_content)

    # Add cells to notebook
    nb.cells = [cell1, cell2, cell3]

    with open(filename, "w") as f:
        nbf.write(nb, f)


def main():
    parser = argparse.ArgumentParser(description="Generate a new Jupyter notebook")
    parser.add_argument("filename", help="Name of the notebook to create")
    parser.add_argument(
        "-db", action="store_true", help="With or without a vector database"
    )
    parser.add_argument(
        "-asnc",
        action="store_true",
        help="Async chat completion instead of a regular one",
    )
    args = parser.parse_args()

    create_notebook(f"{args.filename}.ipynb", args.db, args.asnc)


if __name__ == "__main__":
    main()



import logging

from typing import List, Literal

import tiktoken

from aitemplates.oai.types.base import Message


def num_tokens_from_messages(
    messages: List[Message], model: str = "gpt-3.5-turbo-0301"
) -> int:
    """Returns the number of tokens used by a list of messages.

    Args:
        messages (list): A list of messages, each of which is a dictionary
            containing the role and content of the message.
        model (str): The name of the model to use for tokenization.
            Defaults to "gpt-3.5-turbo-0301".

    Returns:
        int: The number of tokens used by the list of messages."""
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        logging.warning("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if model == "gpt-3.5-turbo":
        logging.warning(
            "Warning: gpt-3.5-turbo may change over time. Returning num tokens assuming gpt-3.5-turbo-0301."
        )
        return num_tokens_from_messages(messages, model="gpt-3.5-turbo-0301")
    elif model == "gpt-4":
        logging.warning(
            "Warning: gpt-4 may change over time. Returning num tokens assuming gpt-4-0314."
        )
        return num_tokens_from_messages(messages, model="gpt-4-0314")
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = (
            4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        )
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif model == "gpt-4-0314":
        tokens_per_message = 3
        tokens_per_name = 1
    else:
        raise NotImplementedError(
            f"""num_tokens_from_messages() is not implemented for model {model}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
        )

    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.raw().items():
            assert isinstance(value, str)
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens



import abc


class Singleton(abc.ABCMeta, type):
    "Singleton metaclass for ensuring only one instance of a class."

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Call method for the singleton metaclass."""
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



from __future__ import annotations

from unittest.mock import patch
import functools
import time
import os
import logging


from openai.error import (
    APIError,
    RateLimitError,
    InvalidRequestError,
    APIConnectionError,
    Timeout,
)

import openai
import openai.api_resources.abstract.engine_api_resource as engine_api_resource
import openai.util
from openai.openai_object import OpenAIObject

from aitemplates.oai.ApiManager import ApiManager


def metered(func):
    """Adds ApiManager metering to functions which make OpenAI API calls.

    Args:
        func (Callable): The function to meter.

    Returns:
        Callable: The metered function.
    """
    api_manager = ApiManager()

    openai_obj_processor = openai.util.convert_to_openai_object

    def update_usage_with_response(response: OpenAIObject):
        """Updates the usage statistics based on the response."""
        try:
            usage = response.usage
            api_manager.update_cost(
                response.usage.prompt_tokens,
                response.usage.completion_tokens if "completion_tokens" in usage else 0,
                response.model,
            )
        except Exception as err:
            logging.error(
                f"Failed to update API costs: {err.__class__.__name__}: {err}"
            )

    def metering_wrapper(*args, **kwargs):
        """Wraps the function in a metering logic."""
        openai_obj = openai_obj_processor(*args, **kwargs)
        if isinstance(openai_obj, OpenAIObject) and "usage" in openai_obj:
            update_usage_with_response(openai_obj)
        return openai_obj

    @functools.wraps(func)
    def metered_func(*args, **kwargs):
        """A function that runs the original function with metering applied."""
        with patch.object(
            engine_api_resource.util,
            "convert_to_openai_object",
            side_effect=metering_wrapper,
        ):
            return func(*args, **kwargs)

    return metered_func


def retry_openai_api(
    num_retries: int = 10,
    backoff_base: float = 2.0,
    warn_user: bool = True,
):
    """Decorate a function to retry OpenAI API call when it fails.

    This function uses exponential backoff strategy for retries.

    Args:
        num_retries (int, optional): Number of retries. Defaults to 10.
        backoff_base (float, optional): Base for exponential backoff. Defaults to 2.0.
        warn_user (bool, optional): Whether to warn the user. Defaults to True.

    Returns:
        Callable: The decorated function.
    """

    def _wrapper(func):
        @functools.wraps(func)
        def _wrapped(*args, **kwargs):
            if "OPENAI_API_KEY" not in os.environ:
                raise ValueError(
                    "OPENAI_API_KEY environment variable must be set when using OpenAI API."
                )
            user_warned = not warn_user
            num_attempts = num_retries + 1  # +1 for the first attempt
            for attempt in range(1, num_attempts + 1):
                try:
                    return func(*args, **kwargs)

                except RateLimitError:
                    if attempt == num_attempts:
                        raise

                    logging.error("Error: Reached rate limit, passing...")
                    if not user_warned:
                        logging.warning(
                            """Please double check that you have setup a paid OpenAI API 
                              Account. You can read more here: https://docs.agpt.co/setup/#getting-an-api-key"""
                        )
                        user_warned = True
                except InvalidRequestError:
                    logging.error("OpenAI API Invalid Request: Prompt was filtered")
                    user_warned = True
                except APIConnectionError:
                    logging.error(
                        "OpenAI API Connection Error: Error Communicating with OpenAI"
                    )
                    user_warned = True
                except Timeout:
                    logging.error("OpenAI APITimeout Error: OpenAI Timeout")
                    user_warned = True
                except APIError as e:
                    if (e.http_status not in [502, 429]) or (attempt == num_attempts):
                        raise

                backoff = backoff_base ** (attempt + 2)
                logging.error(f"Error: API Bad gateway. Waiting {backoff} seconds...")
                time.sleep(backoff)

        return _wrapped

    return _wrapper



from __future__ import annotations

from dataclasses import dataclass, field
import numpy as np
from math import ceil, floor
from typing import Literal, TypedDict, Optional, Union, Any

# Embedding vector
Embedding = Union[list[np.float32], np.ndarray[Any, np.dtype[np.float32]]]

# Token array representing text
TText = list[int]

# Different roles the chat API takes
MessageRole = Literal["system", "user", "assistant"]


class MessageDict(TypedDict):
    role: MessageRole
    content: str


class ResponseDict(TypedDict):
    index: int
    message: MessageDict
    finish_reason: str


class UsageDict(TypedDict):
    prompt_tokens: Optional[int]
    completion_tokens: Optional[int]
    total_tokens: int


@dataclass
class Message:
    """OpenAI Message object containing a role and the message content"""

    role: MessageRole
    content: str

    def raw(self) -> MessageDict:
        return {"role": self.role, "content": self.content}


@dataclass
class ModelInfo:
    """Struct for model information."""

    name: str
    prompt_token_cost: float
    completion_token_cost: float
    max_tokens: int


@dataclass
class ChatModelInfo(ModelInfo):
    "Struct for chat model information."


@dataclass
class EmbeddingModelInfo(ModelInfo):
    "Struct for embedding model information."

    embedding_dimensions: int


@dataclass
class ChatSequence:
    "Utility container for a chat sequence"

    messages: list[Message] = field(default_factory=list)

    def __getitem__(self, i: int):
        return self.messages[i]

    def __iter__(self):
        return iter(self.messages)

    def __len__(self):
        return len(self.messages)

    def __add__(self, other):
        if isinstance(other, ChatSequence):
            # Concatenate the messages lists
            return ChatSequence(self.messages + other.messages)
        else:
            raise TypeError(f"Unsupported operand type for +: {type(other)}")

    def append(self, message: Message):
        return self.messages.append(message)

    def extend(self, messages: list[Message] | ChatSequence):
        return self.messages.extend(messages)

    def insert(self, index: int, *messages: Message):
        for message in reversed(messages):
            self.messages.insert(index, message)

    def add(self, message_role: MessageRole, content: str):
        self.messages.append(Message(message_role, content))

    @property
    def token_length(self):
        from aitemplates.oai.utils.count_tokens import num_tokens_from_messages

        return num_tokens_from_messages(self.messages, self.model.name)

    def raw(self) -> list[MessageDict]:
        return [m.raw() for m in self.messages]

    def expand(self) -> list[Message]:
        return [m for m in self.messages]

    def dump(self) -> str:
        "Return all information stored in the dataclass as a string"
        SEPARATOR_LENGTH = 42

        def separator(text: str):
            half_sep_len = (SEPARATOR_LENGTH - 2 - len(text)) / 2
            return f"{floor(half_sep_len)*'-'} {text.upper()} {ceil(half_sep_len)*'-'}"

        formatted_messages = "\n".join(
            [f"{separator(m.role)}\n{m.content}" for m in self.messages]
        )
        return f"""
        ============== ChatSequence ==============
        Length: {self.token_length} tokens; {len(self.messages)} messages
        {formatted_messages}
        ==========================================
        """

    @staticmethod
    def from_dict(data: dict[str, Any]) -> ChatSequence:
        "Create a chat messages object from a dictionary."
        return ChatSequence(
            [Message(role=x["role"], content=x["content"]) for x in data["messages"]]
        )

    def to_dict(self) -> dict[str, Any]:
        "Convert a chat messages object to a dictionary."
        return {
            "messages": [
                {
                    "role": x.role,
                    "content": x.content,
                }
                for x in self.messages
            ]
        }


@dataclass(frozen=True)
class ChatMessages:
    """A set of prompts to ask.

    Attributes:
        chat_sequences: A set of message sequences, each represents a prompt
    """

    chat_sequences: list[ChatSequence]
    sequential: bool = False

    def __getitem__(self, i: int):
        return self.chat_sequences[i]

    def __iter__(self):
        return iter(self.chat_sequences)

    def __len__(self):
        return len(self.chat_sequences)

    def append(self, message: ChatSequence):
        return self.chat_sequences.append(message)

    def extend(self, messages: list[ChatSequence] | ChatMessages):
        return self.chat_sequences.extend(messages)
